using UnityEngine;
using System.Collections;

/// <summary>
/// Generic Crack
/// </summary>
public class Crack : MonoBehaviour 
{	
	public enum TCrack
	{
		/// <summary>
		/// None crack
		/// </summary>
		NONE_CRACK = 0,
		
		/// <summary>
		/// (first floor) Sewer crack, only one sim per sewer 
		/// Disable after use
		/// Interactive
		/// </summary>
		SEWER_CRACK,
		
		/// <summary>
		/// (first floor) Several sims per hole
		/// Open
		/// </summary>
		HOLE_CRACK,
		
		/// <summary>
		/// (Bunker) Grille crack. Several sims 
		/// Close in first place
		/// Explode when cat is touch
		/// </summary>
		GRILLE_CRACK,
		
		/// <summary>
		/// (Bunker) Trap door crack. Several sims 
		/// Close in first place
		/// Interactive
		/// </summary>
		TRAPDOOR_CRACK,
		
		/// <summary>
		/// (Bunker) Crack door crack. Several sims 
		/// Close in first place
		/// Not interactive
		/// </summary>
		CRACK_CRACK,
		
		/// <summary>
		/// (Bunker) Air ventilation. Several sims 
		/// Close in first place
		/// Not interactive
		/// </summary>
		AIR_CRACK,
		
		/// <summary>
		/// Same as Hole crack but in bunker
		/// </summary>
		HOLE_BUNKER_CRACK,		
		
		/// <summary>
		/// Only for 0 floor
		/// Open
		/// Not interactive
		/// </summary>
		EXIT_CRACK,
		
	}
	
	public Crack.TCrack TypeCrack;
	
	/// <summary>
	/// Number of floor
	/// </summary>
	private int numberFloor = 0;
	public int NumberFloor
	{
		get { return numberFloor; }
		set { numberFloor = value; }
	}
	
	/// <summary>
	/// If crack is interactive
	/// </summary>
	private bool isInteractive = false;
	
	private GameObject interactiveIcon;
	private SkeletonAnimation sklAnimIcon;
	
	/// <summary>
	/// Distance fall for sims
	/// </summary>
	private float distanceFall = 0.0f;		

	/// <summary>
	/// List elements visible in first time
	/// </summary>
	public GameObject[] ListVisibleElements;
	/// <summary>
	/// List elements invisible in first time
	/// </summary>
	public GameObject[] ListInvisibleElements;
	
	/// <summary>
	/// Whether is enableCrack or not
	/// </summary>
	private bool enableCrack;
	
	/// <summary>
	/// Box collider
	/// </summary>
	private BoxCollider triggerCrack;
	
	private bool isAnimated = false;
	/// <summary>
	/// Animation for crack
	/// </summary>
	protected SkeletonAnimation sklAnimCrack;
	
	/// <summary>
	/// Wheter is open or not
	/// </summary>
	private bool isOpen;
	public bool IsOpen
	{
		get { return isOpen; }
		set { isOpen = value; }
	}
	
	/// <summary>
	/// Members for animation
	/// </summary>
	private string nameAnim;
	private float timeAnim = 0.0f;
	private float timeOpen = 0.0f;
	private float updateTAnim = 0.0f;	
	private bool isAnimating;


	/// <summary>
	/// Offsets to apply to height for fall in crack
	/// </summary>
	public float OffsetUpHeight;
	public float OffsetDownHeight;
	/// <summary>
	/// Heights to check if sim is collider with crack
	/// </summary>
	private float upperHeightSim;
	private float bottomHeightSim;
	
	/// <summary>
	/// Whether crack is open at start (debug)
	/// </summary>
	public bool StartOpen = false;

    /// <summary>
    /// If particles are enable
    /// </summary>
    private bool enableParticles = false;
	/// <summary>
	/// Prefab for explode crack
	/// </summary>
	public GameObject PrefabExplodeParticles;
    /// <summary>
    /// Instance of particles for explode crack
    /// </summary>
    private GameObject instExplodeP;
	
	/// <summary>
	/// Transform to exit
	/// </summary>
	public Transform TransformExit;

	/// <summary>
	/// Sound for open crack
	/// </summary>
	public AudioClip OpenCrackSound;

	/// <summary>
	/// Sound explode crack
	/// </summary>
	public AudioClip ExplodeCrackSound;


	void Start ()
	{
		// Set trigger
		this.triggerCrack = this.transform.GetComponent<BoxCollider>();
		this.updateTAnim = 0.0f;
		this.isAnimating = false;
		
		string nameChild = "";
		switch(TypeCrack)
		{
		    case TCrack.SEWER_CRACK:
			    nameChild = "sewerObject";
			    this.isOpen = false;
			    this.isInteractive = true;
			    this.nameAnim = "open";
			    this.timeAnim = 0.83f;
			    this.timeOpen = this.timeAnim/3;
			    this.distanceFall = 200.0f;
			    this.isAnimated = true;
			    this.numberFloor = 0;
			    this.upperHeightSim = this.transform.position.y;
				this.bottomHeightSim = this.transform.position.y - 80.0f ;
			break;
			
		    case TCrack.HOLE_CRACK:			
			    this.isAnimated = false;
			    this.distanceFall =  200.0f;
			    this.isInteractive = false;
			    this.isOpen = true;
			    this.numberFloor = 0;
                this.upperHeightSim = this.transform.position.y + this.OffsetUpHeight;
			    this.bottomHeightSim = this.transform.position.y + this.OffsetDownHeight;
			break;
			
		    case TCrack.HOLE_BUNKER_CRACK:			
			    this.isAnimated = false;
			    this.distanceFall = GameManager.Instance.SimSetupSettings.FloorDistance;
			    this.isInteractive = false;
			    this.isOpen = true;
			    this.numberFloor = 1;
			    this.upperHeightSim = this.transform.position.y + 80.0f;
			    this.bottomHeightSim = this.transform.position.y - 100.0f;
			break;
			
		    case TCrack.TRAPDOOR_CRACK:
			    nameChild = "trapDoorClose";
			    this.isOpen = false;
			    this.isInteractive = true;
			    this.timeOpen = 0.2f;
			    this.timeAnim = timeOpen;
			    this.distanceFall = GameManager.Instance.SimSetupSettings.FloorDistance;
			    this.isAnimated = false;
			    this.upperHeightSim = this.transform.position.y + this.OffsetUpHeight;
			    this.bottomHeightSim = this.transform.position.y + this.OffsetDownHeight;
			    this.numberFloor = 1;
			break;
			
		    case TCrack.CRACK_CRACK:
			    nameChild = "crackCloseObject";
			    this.isOpen = false;
			    this.isInteractive = false;
			    this.distanceFall = GameManager.Instance.SimSetupSettings.FloorDistance;
			    this.isAnimated = false;
			    this.timeAnim = 0.2f;
			    this.timeOpen = 0.2f;
			    this.numberFloor = 1;
			    this.upperHeightSim = this.transform.position.y;
			    this.bottomHeightSim = this.transform.position.y - 70.0f;
			break;
			
		    case TCrack.AIR_CRACK:
			    nameChild = "air";
			    this.isOpen = false;
			    this.isInteractive = false;
			    this.timeAnim = 0.2f;
			    this.timeOpen = this.timeAnim;
			    this.distanceFall = GameManager.Instance.SimSetupSettings.FloorDistance;
			    this.isAnimated = false;
			    this.numberFloor = 	1;
			    this.upperHeightSim = this.transform.position.y + 17.0f;
			    this.bottomHeightSim = this.transform.position.y - 60.0f;
			break;

		    case TCrack.EXIT_CRACK:			
			    this.isAnimated = false;
			    this.distanceFall =  200.0f;
			    this.isInteractive = false;
			    this.isOpen = true;
			    this.numberFloor = 	0;
                this.upperHeightSim = this.transform.position.y + this.OffsetUpHeight;
			    this.bottomHeightSim = this.transform.position.y + this.OffsetDownHeight;			
			break;
			
		}
		
		if (this.isAnimated)
		{		
			Transform auxObject =  this.transform.FindChild(nameChild);
			if (auxObject != null)
			{
				this.sklAnimCrack = auxObject.transform.GetComponentInChildren<SkeletonAnimation>();
			}
		}
		
		if (this.isInteractive)
		{
			this.interactiveIcon = this.transform.FindChild("IntercativeIcon").gameObject;
			if (this.interactiveIcon != null)
			{
				this.sklAnimIcon = interactiveIcon.transform.GetComponentInChildren<SkeletonAnimation>();
			}
		}		
		// If start open is checked open crack
		if (this.StartOpen)
		{
			this.instantOpen();
		}
		
	}
	
	void Update () 
	{
		if (this.isAnimating)
		{
			this.updateTAnim += Time.deltaTime;
			if (this.updateTAnim >= this.timeAnim)
			{
				this.enableIcon(false);
				this.isAnimating = false;			
				
			}
			if (this.updateTAnim >= this.timeOpen)
			{
				this.isOpen = true;
			}
		}

	}
	
	private bool checkHeighSim(float heightSim)
	{
		if ((heightSim >= this.bottomHeightSim) &&
		    (heightSim <= this.upperHeightSim))
		{
			return true;
		}
		return false;
		
	}
	
	void OnTriggerStay(Collider other) 
	{
		if ((this.isOpen) && (other.tag == "Sim"))
		{
			SimController auxSim = other.transform.GetComponent<SimController>();
			if (auxSim != null) 
			{
				if (auxSim.Stats.CurrentState != SimController.State.SIM_FALL)
				{
					bool matchHeight = checkHeighSim(auxSim.transform.position.y);
					if (matchHeight)
					{
						GameObject shadow = other.transform.FindChild("bodyColliderShadow").gameObject;
						if (shadow != null)
						{
							switch(TypeCrack)
							{
								case TCrack.SEWER_CRACK:
									shadow.SetActive(false);	
								break;
							}
						}
						
						// Set sim behid foreground
						float auxPosZSim = this.transform.position.z - 0.05f;
						other.transform.position = new Vector3(this.transform.position.x,
						                                       other.transform.position.y, auxPosZSim);
						
						auxSim.SetupSimFall.DistanceFall = this.distanceFall;
						auxSim.SetupSimFall.CurrentFloor = auxSim.Stats.Number_Floor;
						auxSim.SetupSimFall.IsOutsideBunker = false;
                        // Check if is zombie and is attacking another sim
                        if (auxSim.Stats.CurrentState == SimController.State.Z_ATTACK)
                        {
                            if (auxSim.Stats.SimChase != null)
                            {
                                                       
                                 // If my sim chase is on infected zombie and is me, then change elements
                                if (auxSim.Stats.SimChase.Stats.CurrentState == SimController.State.INFECT_ZOMBIE)
                                {
                                    auxSim.Stats.SimChase.Stats.IsInfecting = false;
                                    auxSim.Stats.SimChase.DestroyInfectedBar();

                                    if (auxSim.transform.position.x <= auxSim.Stats.SimChase.transform.position.x)
                                    {
                                        auxSim.Stats.SimChase.SetupAlertSim.DirectionWarn = SimController.DirectionType.RIGHT;
                                    }
                                    else
                                    {
                                        auxSim.Stats.SimChase.SetupAlertSim.DirectionWarn = SimController.DirectionType.LEFT;
                                    }

                                    switch (auxSim.Stats.SimChase.Stats.TSkin)
                                    {
                                        case SimController.Skin.BUSINESS_SKIN:
                                        case SimController.Skin.GEEK_SKIN:
                                        case SimController.Skin.MOBILE_SKIN:
                                        case SimController.Skin.MUSIC_SKIN:
                                            auxSim.Stats.SimChase.ChangeState(SimController.State.C_WARN);
                                            break;

                                        case SimController.Skin.POLY_SKIN:
                                            auxSim.Stats.SimChase.ChangeState(SimController.State.C_ALERT);
                                        break;
                                        case SimController.Skin.NONE_SKIN:
                                            auxSim.Stats.SimChase.ChangeState(SimController.State.C_IDLE);
                                        break;

                                    }
                                    // If my sim to chase, have sim hunted and is me, then take of that sim hunted
                                    if (auxSim.Stats.SimChase.Stats.SimHunted != null)
                                    {
                                        if (auxSim.Stats.SimChase.Stats.SimHunted.Stats.IDSim == auxSim.Stats.IDSim)
                                        {
                                            auxSim.Stats.SimChase.Stats.SimHunted = null;
                                        }
                                    }
                                    auxSim.Stats.SimChase = null;
                                }
                               
                            }
                        }

						auxSim.ChangeState(SimController.State.SIM_FALL);
						
						
						
						switch(TypeCrack)
						{
						    case TCrack.SEWER_CRACK:
							    //this.triggerCrack.enabled = false;
							    //this.enabled = false;
							break;
						}
					}
				}					
			}
		}
	}

    public void changeAnim(string nameAnim)
    {
        if (this.sklAnimCrack != null)
        {
            this.sklAnimCrack.animationName = nameAnim;
        }	
    }

	public void playAnim()
	{
		if (this.sklAnimCrack != null)
		{
			this.sklAnimCrack.animationName = this.nameAnim;
			this.isAnimating = true;			
		}	
	}

	public void instantOpen()
	{
		switch(TypeCrack)
		{
			case TCrack.SEWER_CRACK:
                this.changeAnim("open2");
			break;
				
			case TCrack.TRAPDOOR_CRACK:
                this.swapElements();

			break;
			case TCrack.CRACK_CRACK:
                this.swapElements();

			break;
			case TCrack.AIR_CRACK:
                this.swapElements();

			break;
			
		}
		this.hideIcon();
        this.isOpen = true;
	}

	
	public void Open()
	{
		switch(TypeCrack)
		{
		case TCrack.SEWER_CRACK:
			this.playAnim();
			this.enableIcon(true);
			GameManager.Instance.Sound_Manager.PlaySoundEffect(this.OpenCrackSound);
		break;
			
		case TCrack.TRAPDOOR_CRACK:
			this.explodeCrack();
			this.enableIcon(true);
			GameManager.Instance.Sound_Manager.PlaySoundEffect(this.OpenCrackSound);
		break;
		case TCrack.CRACK_CRACK:
			this.explodeCrack();
			GameManager.Instance.Sound_Manager.PlaySoundEffect(this.ExplodeCrackSound);
		break;
		case TCrack.AIR_CRACK:
			this.explodeCrack();
			GameManager.Instance.Sound_Manager.PlaySoundEffect(this.ExplodeCrackSound);
		break;
			
		}
		
	}
	/// <summary>
	/// Methos to explode crack
	/// </summary>
	private void explodeCrack()
	{
        // Particles
        if (this.enableParticles)
        {
            if (this.PrefabExplodeParticles != null)
            {
                Vector3 auxPos = new Vector3(this.transform.position.x,
                                             this.transform.position.y - 20.0f,
                                             this.transform.position.z - 0.2f);
                this.instExplodeP = Instantiate(this.PrefabExplodeParticles, auxPos, Quaternion.identity) as GameObject;
                Invoke("destroyExplodeParticle", 0.5f);
            }
        }
		
		this.swapElements();
		this.isAnimating = true;

	}

	private void swapElements()
	{
		if (this.ListVisibleElements != null)
		{
			foreach (GameObject aux_v in this.ListVisibleElements)
			{
				aux_v.SetActive(false);
			}
		}
		
		if (this.ListInvisibleElements != null)
		{
			foreach (GameObject aux_v in this.ListInvisibleElements)
			{
				aux_v.SetActive(true);
			}
		}
	}

    private void destroyExplodeParticle()
    {
        if (this.instExplodeP != null)
        {
            Destroy(this.instExplodeP);
        }
    }
	
	public void OnItemClick()
	{	
		if ((this.isInteractive) && (!this.isOpen))
		{	
			this.Open();
		}
	}	

	private void hideIcon()
	{
		if (this.isInteractive)
		{
			if (this.interactiveIcon != null)
			{
				this.interactiveIcon.SetActive(false);
			}
		}
	}

	private void enableIcon(bool activeIcon)
	{
		if (this.isInteractive)
		{
			if (this.interactiveIcon != null)
			{
				if (activeIcon)
				{
					this.sklAnimIcon.animationName = "activate";
				}else
				{
					this.sklAnimIcon.animationName = "deactivate";
				}		
			}
		}		
	}
}
